from __future__ import annotations

from typing import Dict, Tuple, Optional, List, Set

from .analysis import Analysis
from lang.cfg import Entry, Exit, InnerNode, Node
from lang.ast import Assignment, Guard, Variable, Expression
from lang.utils import getVarsUsedInStatement

AnalysisType = Set[Expression]


class AvailableExpressions(Analysis[AnalysisType]):
    T = AnalysisType
    def initial_in(self, node: Node) -> T:
        return set()

    def initial_out(self, node: Node) -> T:
        return set()

    def get_new_values(self, node: Node, predecessors_out: List[T], successors_in: List[T],
                       curr_out: Dict[Node, T], curr_in: Dict[Node, T]) -> Tuple[T, T]:

        Kill: AnalysisType = set()
        Gen: AnalysisType = set()

        In: AnalysisType = set().union(*predecessors_out)

        if isinstance(node, Entry):
            ...
        elif isinstance(node, Exit):
            ...
        elif isinstance(node, InnerNode):
            if isinstance(node.statement, Guard):
                Gen = {node.statement.expression}
            elif isinstance(node.statement, Assignment):
                a = node.statement.value
                x = node.statement.variable
                # is x used in rhs
                Kill = {a} if x in getVarsUsedInStatement(node.statement) else {}
                Gen = {a} if x not in getVarsUsedInStatement(node.statement) else {}
            else:
                raise Exception(
                    f"InnerNode {node} with statement of type {type(node.statement)} not supported in {type(self).__name__}'s get_new_values method.")
        else:
            raise Exception(
                f"Node {node} of type {type(node)} not supported in {type(self).__name__}'s get_new_values method.")

        Out: AnalysisType = Gen.union(In.difference(Kill))
        return (In, Out)
