from __future__ import annotations

from typing import Dict, List, Set, Tuple

from .analysis import Analysis
from lang.cfg import Entry, Exit, InnerNode, Node
from lang.ast import Assignment, Guard, Variable
from lang import getVarsUsedInStatement

AnalysisType = Set[Variable]


class LiveVariables(Analysis[AnalysisType]):
    T = AnalysisType

    def initial_in(self, node: Node) -> T:
        return set()

    def initial_out(self, node: Node) -> T:
        return set()

    def get_new_values(self, node: Node, predecessors_out: List[T], successors_in: List[T],
                       curr_out: Dict[Node, T], curr_in: Dict[Node, T]) -> Tuple[T, T]:

        """ 
        Return tuple (In, Out), computed using this node's predessors' Outs (`predecessors_out`) and/or 
        its succesors' Ins (`successors_in`). 
        """

        Gen = set()
        Kill = set()

        # Update Gen and Kill, depending on Node type 
        if isinstance(node, Entry):
            pass
        elif isinstance(node, Exit):
            pass
        elif isinstance(node, InnerNode):
            if isinstance(node.statement, Guard):
                Gen = getVarsUsedInStatement(node.statement)
            elif isinstance(node.statement, Assignment):
                Gen = getVarsUsedInStatement(node.statement)
                Kill = {node.statement.variable}
            else:
                raise Exception(
                    f"InnerNode {node} with statement of type {type(node.statement)} not supported in {type(self).__name__}'s get_new_values method.")
        else:
            raise Exception(
                f"Node {node} of type {type(node)} not supported in {type(self).__name__}'s get_new_values method.")

        Out = set().union(*successors_in)
        In = Out.difference(Kill).union(Gen)

        return (In, Out)
