from __future__ import annotations

from typing import Dict, Tuple, Optional, List, Set

from .analysis import Analysis
from lang.cfg import Entry, Exit, InnerNode, Node
from lang.ast import Assignment, Guard, Variable

AnalysisType = Set[Tuple[Variable, Optional[Node]]]


class ReachingDefinitions(Analysis[AnalysisType]):
    T = AnalysisType
    def initial_in(self, node: Node) -> AnalysisType:
        return set()

    def initial_out(self, node: Node) -> AnalysisType:
        return set()

    def get_new_values(self, node: Node, predecessors_out: List[AnalysisType], successors_in: List[AnalysisType],
                       curr_out: Dict[Node, AnalysisType], curr_in: Dict[Node, AnalysisType]) -> Tuple[
        AnalysisType, AnalysisType]:

        Kill: AnalysisType = set()
        Gen: AnalysisType = set()

        In: AnalysisType = set().union(*predecessors_out)

        if isinstance(node, Entry):
            ...
        elif isinstance(node, Exit):
            ...
        elif isinstance(node, InnerNode):
            if isinstance(node.statement, Guard):
                ...
            elif isinstance(node.statement, Assignment):
                var = node.statement.variable
                Kill = set(filter(lambda tup: tup[0].id == var.id, In))
                Gen = {(var, node)}
            else:
                raise Exception(
                    f"InnerNode {node} with statement of type {type(node.statement)} not supported in {type(self).__name__}'s get_new_values method.")
        else:
            raise Exception(
                f"Node {node} of type {type(node)} not supported in {type(self).__name__}'s get_new_values method.")

        Out: AnalysisType = Gen.union(In.difference(Kill))
        return (In, Out)
