from lang.ast import AST, BinOp, BinaryExpression, BooleanConstant, Expression, Assignment, Branch, IntegerConstant, Statement, UnOp, UnaryExpression, Variable, WhileLoop
from lark import Transformer, v_args


class ToAST(Transformer):

    variables: set[Variable]

    def __init__(self, visit_tokens: bool = True) -> None:
        super().__init__(visit_tokens)
        self.variables = set() 

    @v_args(inline=True)
    def program(self, statements):
        return AST(statements, self.variables)

    def statements(self, lst):
        return lst
    
    def statement(self, lst):
        assert(len(lst) == 2)
        assert(isinstance(lst[0], str))
        assert(isinstance(lst[1], (Assignment, WhileLoop, Branch)))
        label = lst[0]
        lst[1].label = label 
        return lst[1]
    
    @v_args(inline=True)
    def label(self, s: str): 
        assert(isinstance(s, str))
        return s

    @v_args(inline=True)
    def assignment(self, var: Variable, expression: Expression):
        assert(isinstance(var, Variable))
        # assert(isinstance(expression, Expression))
        return Assignment(var, expression, None)

    @v_args(inline=True)
    def condition(self, cond: Expression, br1: list[Statement], br2: list[Statement]):
        return Branch(cond, br1, br2, None)

    @v_args(inline=True)
    def while_loop(self, cond: Expression, stmts: list[Statement]):
        return WhileLoop(cond, stmts, None)

    @v_args(inline=True)
    def bin_expression(self, lhs: Expression, op_str: str, rhs: Expression):
        assert(isinstance(op_str, str))
        # assert(isinstance(lhs, Expression))
        # assert(isinstance(rhs, Expression))
        if op_str == "+":
            op = BinOp.Plus
        elif op_str == "-":
            op = BinOp.Minus
        elif op_str == "*":
            op = BinOp.Times
        elif op_str == "/":
            op = BinOp.Div
        elif op_str == "||":
            op = BinOp.Or
        elif op_str == "&&":
            op = BinOp.And
        elif op_str == "<":
            op = BinOp.LessThan
        elif op_str == "<=":
            op = BinOp.LessThanEquals
        elif op_str == ">":
            op = BinOp.GreaterThan
        elif op_str == ">=":
            op = BinOp.GreaterThanEquals
        elif op_str == "==":
            op = BinOp.Equals
        elif op_str == "!=":
            op = BinOp.NotEquals
        return BinaryExpression(lhs, op, rhs)

# un_expression: unop expression
    @v_args(inline=True)
    def un_expression(self, op_str: str, expression: Expression):
        assert(isinstance(op_str, str))
        # assert(isinstance(expression, Expression))
        if op_str == "-":
            op = UnOp.Minus
        elif op_str == "!":
            op = UnOp.Not
        return UnaryExpression(op, expression)

# unop: "-" | "!"

    @v_args(inline=True)
    def int_const(self, s: int):
        assert(isinstance(s, int))
        return IntegerConstant(s)

    @v_args(inline=True)
    def bool_const(self, s):
        assert(isinstance(s, str))
        if s == "true":
            return BooleanConstant(True)
        elif s == "false":
            return BooleanConstant(False)

    @v_args(inline=True)
    def var(self, val: str):
        var = Variable(val)
        self.variables.add(var)
        return var

    def ID(self, s):
        return str(s)

    def INTEGER(self, n):
        return int(n)
