from analysis.fixed_point import fixed_point
from analysis.reaching_definitions import ReachingDefinitions
from lang.ast.ast import Assignment, IntegerConstant, Variable
from lang.to_cfg import stringToCFG
from lang.utils import getExitValue, getNodeByLabel, getValueByLabel

x_eq_0 = Assignment(Variable("x"), IntegerConstant(0), None)


def test_reaching_definitions_sequential1():
    s = """
    l1: x = 0; 
    l2: y = 10;  
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, ReachingDefinitions(cfg))
    assert(getValueByLabel("l2", In) == {
           (Variable("x"), getNodeByLabel("l1", cfg.nodes))})
    assert(getValueByLabel("l2", Out) == {(Variable("x"), getNodeByLabel(
        "l1", cfg.nodes)), (Variable("y"), getNodeByLabel("l2", cfg.nodes))})


def test_reaching_definitions_sequential2():
    s = """
    l1: x = 0; 
    l2: x = 10;  
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, ReachingDefinitions(cfg))
    assert(getValueByLabel("l2", In) == {
           (Variable("x"), getNodeByLabel("l1", cfg.nodes))})
    assert(getValueByLabel("l2", Out) == {
           (Variable("x"), getNodeByLabel("l2", cfg.nodes))})


def test_reaching_definitions_branch1():
    s = """
    l1: x = 0; 
    l2: y = 0; 
    l3: if (x > 0) { 
    l4:     x = 1; 
        } else {
    l5:     y = 1; 
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, ReachingDefinitions(cfg))
    assert(getExitValue(In) == {
           (Variable("x"), getNodeByLabel("l1", cfg.nodes)),  (Variable("y"), getNodeByLabel("l2", cfg.nodes)),  (Variable("x"), getNodeByLabel("l4", cfg.nodes)),  (Variable("y"), getNodeByLabel("l5", cfg.nodes))})


def test_reaching_definitions_branch2():
    s = """
    l1: x = 0; 
    l2: y = 0; 
    l3: if (x > 0) { 
    l4:     y = 1; 
        } else {
    l5:     y = 2; 
        }
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, ReachingDefinitions(cfg))
    assert(getExitValue(In) == {
           (Variable("x"), getNodeByLabel("l1", cfg.nodes)), (Variable("y"), getNodeByLabel("l4", cfg.nodes)),  (Variable("y"), getNodeByLabel("l5", cfg.nodes))})


def test_reaching_definitions_loop1():
    s = """
    l1: x = 0; 
    l2: y = 0; 
    l3: while (x > 0) { 
    l4:     x = 1; 
    l5:     y = 1; 
        }
    l6: y = 2; 
    """
    cfg = stringToCFG(s)
    In, Out = fixed_point(cfg, ReachingDefinitions(cfg))
    assert(getValueByLabel("l3", In) == {
           (Variable("x"), getNodeByLabel("l1", cfg.nodes)), (Variable("y"), getNodeByLabel("l2", cfg.nodes)), (Variable("x"), getNodeByLabel("l4", cfg.nodes)),  (Variable("y"), getNodeByLabel("l5", cfg.nodes))})
    assert(getExitValue(In) == {
           (Variable("x"), getNodeByLabel("l1", cfg.nodes)), (Variable("x"), getNodeByLabel("l4", cfg.nodes)), (Variable("y"), getNodeByLabel("l6", cfg.nodes))})
